const http = require('http');
const { Semantics, Semantic } = require('./src/semantics');
const { Generators, Generator } = require('./src/generators');
const { Config } = require('./src/config');
const fetch = require('node-fetch')
const _ = require('underscore');
const base64 = require('base-64')

const processContexts = (contexts, params) => {
  contextsPrime = []
  generateds = []
  for (context of contexts) {
    const result = processContext(context, params)
    contextsPrime.push(result.context)
    generateds.push(result.generated)
  }
  return { contexts: contextsPrime, generated: generateds }
}

const processContext = (context, { objects = {}, generators = [], semantics = [] }) => {
  generators = new Generators(generators.map( (g) => new Generator(g[0], g[1]) ) )
  semantics = new Semantics(semantics.map( (g) => new Semantic(g[0], g[1]) ) )

  context = semantics.apply(objects, context)
  const generated = generators.apply(objects, context)[0]

  return { context, generated }
}

const process = (url, key, config, retries = 2) => {
  let data = Object.assign( { key, version: '3' }, config.config );
  const generators = new Generators(data.generators.map( (g) => new Generator(g[0], g[1]) ) )
  delete data['generators']
  const semantics = new Semantics(data.semantics.map( (g) => new Semantic(g[0], g[1]) ) )
  delete data['semantics']

  const doWithRetries = (n) => {
    return fetch(`${url}/process`, {
      method: "POST",
      body: JSON.stringify(data),
      timeout: 1000 * 60 * 5, // it does not respect this timeout so that's why I have the retries
      headers: {
        mode: "no-cors",
        "Content-Type": "application/json"
      },
      }).then( result => {
        if (result.ok) {
          return result.json()
        }
        if (result.status == 504){
          if (n == 0) {
            throw `Error ${result.status} - ${result.statusText}`;
          } else {
            return doWithRetries(n-1)
          }
        }
        if (result.status == 503) {
          throw `Error ${result.status} - ${result.statusText}. If you are using the developer version try restarting your entodicton server. If takes a few minutes to be up.`;
        } else {
          throw `Error ${result.status} - ${result.statusText}`;
        }
      })
    }

  let todo = doWithRetries(retries);
  return todo
    .then( json => {
      if (json.status != 200) {
        throw json;
      }
      if (json.results) {
        let resultsPrime = []
        let generatedsPrime = []
        for (let i = 0; i < json.results.length; ++i ) {
          let resultPrime = []
          let generatedPrime = []
          const result = json.results[i];
          for (let j = 0; j < result.length; ++j) {
            const context = result[j];
            let contextPrime = context
            if (!config.get('skipSemantics')) {
              contextPrime = semantics.apply(config.config.objects, context)
            }
            const generated = generators.apply(config.config.objects, contextPrime)[0]
            resultPrime.push(contextPrime)
            generatedPrime.push(generated)
          }
          resultsPrime.push(resultPrime);
          generatedsPrime.push(generatedPrime);
          json.results = resultsPrime;
          json.generated = generatedsPrime;
        }
      }
      if (json.contexts) {
        json.results = json.results[0]
        json.generated = json.generated[0]
      }
      return json;
    });
}

const matching = (actual, expected) => {
  if (!_.isEqual(actual, expected)) {
    return false;
  }
  return true;
}

const submitBugToAPI = (subscription_id, subscription_password, config) => {
  console.log("********* Submitting bug *********")
  const body = { description: config.config.description, config: config.config };
  //fetch('http://localhost:5000/bug', {
  fetch('https://thinktelligence.com/api/bug', {
      method: "POST",
      body: JSON.stringify(body),
      headers: {
        mode: "no-cors", // Type of mode of the request
        "Content-Type": "application/json", // request content type
        "Authorization": 'Basic ' + base64.encode(subscription_id + ":" + subscription_password)
      } 
    }).then( (result) => result.json() )
      .then( (json) => {
      if (json.error) {
        console.log(`Error submitting the bug: ${json.error}`);
      } else {
        console.log(`New bug id id ${json.id}`);
      }
    });
}

const submitBug = (subscription_id, subscription_password, entodicton_url, entodicton_key, config, retries = 2) => {
  const properties = ['expected_results', 'expected_generated', 'utterances']
  properties.forEach( (property) => {
    if (!config.get('expected_results') ) {
      throw "Missing property expected_results";
    }
  });

  return process(url, key, config)
    .then( (responses) => {
      let hasError = false;
      if (!matching(responses.results, config.config.expected_results)) {
        console.log('JSON does not match');
        console.log('actual', JSON.stringify(responses.results))
        console.log('expected', JSON.stringify(config.config.expected_results))
        hasError = true;
      }
      if (!matching(responses.generated, config.config.expected_generated)) {
        console.log('Generated does not match');
        console.log('actual', JSON.stringify(responses.generated))
        console.log('expected', JSON.stringify(config.config.expected_generated))
        hasError = true;
      }

      if (hasError) {
        submitBugToAPI(subscription_id, subscription_password, config)
      } else {
        throw "\n\n *** The expected results matched the generated results so the bug was not submitted since its working. ***\n\n:"
      }
    })
    .catch( (error) => { 
      console.log('Error', error)
      throw error;
    })
}

module.exports = { process, submitBug, processContext, processContexts }
